/*
 * SDDModule.java
 *
 * Created on June 12, 2006, 12:54 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package sddm;
import java.io.*;
import javax.media.opengl.*;
import java.awt.event.*;
import net.java.joglutils.*;

/**
 * //TODO: write guide to sub-classing
 * Make sure to remember to describe how to subclass arguments
 * must override:
 * void drawGeometry(GL gl)
 * static SDDModuleControlPanel newModuleControlPanel()
 *
 * should override:
 * static String moduleTitle()
 * float getMinFeatureSizeHint()
 * static SDDModuleSetupPanel newStartupPanel()
 *
 * may override:
 * void init(GL gl)
 * void preDraw(GL gl)
 * void postDraw(GL gl) -- make sure to make super call
 * void resize(GL gl, int x, int y, int width, int height)
 * void attachEventListeners(StereoCameraControlFrame frame) --- use frame.add<ListenerType>ToFrames -- make sure to make super call
 * SDDModule makeModuleFromArgs(String modName, java.util.HashMap<String,Integer> switchMap, String[] args, StereoCamera camera)
 *
 * @author Erik J. Tollerud
 */
public abstract class SDDModule implements java.io.Serializable {
    
    /**
     * The {@link StereoCamera} attached to this module
     */
    protected StereoCamera cam;
    /**
     * Indicates if display list needs to be recompiled on the next display
     */
    protected boolean recompileDisplayList; //has package access for SDDMListener
    private boolean displayListOn;
    protected float[] axisSizes;
    protected Integer axisTicks;
    protected Integer axisDispList;
    private float[] bgColorComponents;
    
    /**
     * Creates a new instance of SDDModule with a new camera.  In an overriden module, minFeatureSizeHint should be set.
     */
    public SDDModule() {
        this(new StereoCamera());
    }
    /**
     * Creates a new instance of SDDModule linked to an existing camera. In an overriden module, minFeatureSizeHint should be set.
     * @param camera The camera to attach to this module
     */
    public SDDModule(StereoCamera camera) {
        this.cam = camera;
        recompileDisplayList = true;
        axisSizes = null;
        axisTicks = null;
        axisDispList = null;
        bgColorComponents = new float[4];
        bgColorComponents[3] = 1.0f;
    }
    
    /**
     * override with code that draws the geometry of the module
     * @param gl the OpenGL instance to draw into
     */
    public abstract void drawGeometry(GL gl);
    /**
     * Initialization code called when the module is generated
     * @param gl the OpenGL instance to draw into
     */
    public void init(GL gl) {
    }
    
    /**
     * Code called at beginning or OpenGL redisplay (outside display list)
     * @param gl the OpenGL instance to draw into
     */
    public void preDraw(GL gl) {
    }
    
    /**
     * override with code to run after the geometry is drawn (outside display list)
     * @param gl the OpenGL instance to draw into
     */
    public void postDraw(GL gl) {
        
        //added for modules that clear during drawing phase
        if (cam.drawSphereOnUpdate) {
            cam.sphereSizeOnUpdate = 5.0f*this.getMinFeatureSizeHint();
            cam.drawTargetSphere(gl,cam.sphereSizeOnUpdate);
        }
        
        
        if (axisSizes != null) {
            
            if (axisDispList == null) {
                axisDispList = gl.glGenLists(1);
                gl.glNewList(axisDispList,GL.GL_COMPILE);
                
                gl.glBegin(gl.GL_LINES);
                gl.glColor3f(1,0,0);
                gl.glVertex3f(0,0,0);
                gl.glVertex3f(axisSizes[0],0,0);
                gl.glColor3f(0,1,0);
                gl.glVertex3f(0,0,0);
                gl.glVertex3f(0,axisSizes[1],0);
                gl.glColor3f(0,0,1);
                gl.glVertex3f(0,0,0);
                gl.glVertex3f(0,0,axisSizes[2]);
                
                if (this.axisTicks != null) {
                    //draw Ticks
                    com.sun.opengl.util.GLUT glut = new com.sun.opengl.util.GLUT();
                    float scaledMin = 3*this.getMinFeatureSizeHint();
                    float sqrtTwoScaledMin = scaledMin*(float)Math.sqrt(2.0);
                    float tickStep = 1.0f/(float)axisTicks;
                    final float axisDrawMin = this.getMinFeatureSizeHint()/10; //don't draw axis that are of very small size - makes the thing crap out in the for loop
                    
                    //x-axis tick drawn parallel to y-axis
                    if (axisSizes[0] > axisDrawMin) {
                        gl.glColor3f(1,0,0);
                        for(float f = tickStep*axisSizes[0];f<=axisSizes[0];f+=tickStep*axisSizes[0]) {
                            gl.glVertex3f(f,-scaledMin,0);
                            gl.glVertex3f(f,scaledMin,0);
                        }
                    }
                    
                    //y-axis tick drawn parallel to x-axis
                    if (axisSizes[1] > axisDrawMin) {
                        gl.glColor3f(0,1,0);
                        for(float f = tickStep*axisSizes[1];f<=axisSizes[1];f+=tickStep*axisSizes[1]) {
                            gl.glVertex3f(-scaledMin,f,0);
                            gl.glVertex3f(scaledMin,f,0);
                        }
                    }
                    
                    //z-axis tick drawn diagonally along x=y line
                    if (axisSizes[2] > axisDrawMin) {
                        gl.glColor3f(0,0,1);
                        for(float f = tickStep*axisSizes[2];f<=axisSizes[2];f+=tickStep*axisSizes[2]) {
                            gl.glVertex3f(-sqrtTwoScaledMin,-sqrtTwoScaledMin,f);
                            gl.glVertex3f(sqrtTwoScaledMin,sqrtTwoScaledMin,f);
                        }
                    }
                    gl.glEnd();
                    
                    //draw numbers to go with tick marks
                    for(float f = tickStep;f<=1.0f;f+=tickStep) {
                        java.text.DecimalFormat sciNotForm= new java.text.DecimalFormat();
                        sciNotForm.applyPattern("0.###E0");
                        java.text.DecimalFormat decForm= new java.text.DecimalFormat();
                        
                        if (axisSizes[0] > axisDrawMin) {
                            gl.glColor3f(1,0,0);
                            gl.glRasterPos3f(f*axisSizes[0],0,0);
                            if (f*axisSizes[0] >= 1000)
                                glut.glutBitmapString(glut.BITMAP_HELVETICA_10,sciNotForm.format(f*axisSizes[0]));
                            else
                                glut.glutBitmapString(glut.BITMAP_HELVETICA_10,decForm.format(f*axisSizes[0]));
                        }
                        
                        if (axisSizes[1] > axisDrawMin) {
                            gl.glColor3f(0,1,0);
                            gl.glRasterPos3f(0,f*axisSizes[1],0);
                            if (f*axisSizes[1] >= 1000)
                                glut.glutBitmapString(glut.BITMAP_HELVETICA_10,sciNotForm.format(f*axisSizes[1]));
                            else
                                glut.glutBitmapString(glut.BITMAP_HELVETICA_10,decForm.format(f*axisSizes[1]));
                        }
                        
                        if (axisSizes[2] > axisDrawMin) {
                            gl.glColor3f(0,0,1);
                            gl.glRasterPos3f(0,0,f*axisSizes[2]);
                            if (f*axisSizes[2] >= 1000)
                                glut.glutBitmapString(glut.BITMAP_HELVETICA_10,sciNotForm.format(f*axisSizes[2]));
                            else
                                glut.glutBitmapString(glut.BITMAP_HELVETICA_10,decForm.format(f*axisSizes[2]));
                        }
                    }
                } else
                    gl.glEnd();
                
                gl.glEndList();
            }
            
            gl.glDisable(gl.GL_LINE_STIPPLE);
            gl.glDisable(gl.GL_LIGHTING);
            gl.glLineWidth(1.0f);
            gl.glCallList(axisDispList);
        }
        
    }
    
    /**
     * Called when the GLAutoDrawable that the SDDMListener resides upon is resize
     * @param gl the GL Context for the module
     * @param x the upper-left corner of the viewport's new x-coordinate
     * @param y the upper-left corner of the viewport's new y-coordinate
     * @param width the new x-size of the component
     * @param height the new y-size of the component
     */
    public void  resize(GL gl, int x, int y, int width, int height) {
        
    }
    
    /**
     * generates the title of the module.  Override for new modules
     * @return the title of the module
     */
    public static String moduleTitle() {
        return "Stereo Data Display Module";
    }
    /**
     * generates the title of the module from the runtime class of the provided module. DO NOT override - use moduleTitle() to specify a title.
     * @return the title of the module
     */
    public static String moduleTitle(SDDModule modToGetTitleFrom) {
        String str = "Stereo Data Display Module";
        Class moduleClass = SDDModule.class;
        if (modToGetTitleFrom!=null)
            moduleClass = modToGetTitleFrom.getClass();
        try {
            java.lang.reflect.Method ts = moduleClass.getMethod("moduleTitle");
            str = (String) ts.invoke(null);
        } catch (Exception e) {
            System.err.println("Exception getting name:"+e);
            str = "Could not load title:"+moduleClass.toString();
        }
        return str;
    }
    
    public String toString() {
        return this.moduleTitle();
    }
    
    /**
     * Generates a control panel of the type for this module, and links a new camera to the panel
     * @return a newly instantiated control panel
     */
    public abstract SDDModuleControlPanel newModuleControlPanel();
    
    /**
     * Generates a JPanel for inputting data before the object is instantiated. The size is expected to be 355-by-120 or smaller. Module is responsible for applying any desired settings from the panel.
     * @return a newly instantiated panel, or null if there is no specific setup panel for that module
     */
    
    public static SDDModuleSetupPanel newStartupPanel() {
        return null;
    }
    
    /**
     *
     * @param camera
     */
    public void setCamera(StereoCamera camera) {
        this.cam = camera;
    }
    
    /**
     * Returns the camera used by this module
     * @return  the camera used in this module
     */
    public StereoCamera getCamera() {
        return cam;
    }
    
    //Event Listeners
    /**
     * Attaches module-specific event listeners to the specified frame.  If overriden, use frame.add<ListenerType>ToFrames is recommended
     * @param frame the control frame connected to the OpenGL windows
     */
    public void attachEventListeners(StereoCameraControlFrame frame) {
        SDDMMouseListener ml = new SDDMMouseListener(frame);
        frame.addKeyListenerToJFrames(new SDDMKeyListener(frame));
        frame.addMouseListenerToJFrames(ml);
        frame.addMouseMotionListenerToJFrames(ml);
        frame.addMouseWheelListenerToJFrames(ml);
    }
    
    public void setDisplayListOn(boolean displayListOn) {
        this.displayListOn = displayListOn;
    }
    
    public boolean isDisplayListOn() {
        return this.displayListOn;
    }
    
    public void setBackgroundColor(java.awt.Color background) {
        bgColorComponents = background.getComponents(bgColorComponents);
    }
    
    public java.awt.Color getBackgroundColor() {
        return new java.awt.Color(bgColorComponents[0],bgColorComponents[1],bgColorComponents[2],bgColorComponents[3]);
    }
    
    void applyClearColor(GL gl) {
        gl.glClearColor(bgColorComponents[0],bgColorComponents[1],bgColorComponents[2],bgColorComponents[3]);
    }
    
    public void recompileDisplayList() {
        this.recompileDisplayList = true;
    }
    
    /**
     * Determnies if the x,y, and z axis will be drawn as red, green, and blue colors with no values being drawn.
     *
     * @param drawAxis true if the axis should be drawn.  If false, other arguments are ignored
     * @param xSize the size of the x-axis
     * @param ySize the size of the y-axis
     * @param zSize the size of the z-axis
     */
    public void setDrawAxis(boolean drawAxis, float xSize, float ySize, float zSize) {
        this.setDrawAxis(drawAxis,xSize,ySize,zSize,5);
    }
    
    /**
     * Determnies if the x,y, and z axis will be drawn as red, green, and blue Colors.
     *
     *
     * @param numberOfTicks number of marks to draw along axis
     * @param drawAxis true if the axis should be drawn.  If false, other arguments are ignored
     * @param xSize the size of the x-axis
     * @param ySize the size of the y-axis
     * @param zSize the size of the z-axis
     */
    public void setDrawAxis(boolean drawAxis, float xSize, float ySize, float zSize, int numberOfTicks) {
        if (drawAxis) {
            this.axisSizes = new float[3];
            axisSizes[0] = xSize;
            axisSizes[1] = ySize;
            axisSizes[2] = zSize;
        } else
            this.axisSizes = null;
        
        if (numberOfTicks > 0)
            axisTicks = numberOfTicks;
        else
            axisTicks = null;
        
        axisDispList = null;
        
    }
    
    public boolean isDrawAxis() {
        return (axisSizes != null);
    }
    
    public int getNumOfTicks() {
        if(this.axisTicks == null)
            return 0;
        else
            return this.axisTicks;
    }
    
    /**
     * Gives some clue as to the size of individual elements of this module. Override to specify the hint for this module.
     * @return the size of the smallest feature in this module
     */
    public float getMinFeatureSizeHint() {
        return 1.0f;
    }
    
    public void configurePanels(StereoCameraControlPanel sccp, SDDModuleControlPanel smcp) {
        sccp.setTransStep(10.0f*getMinFeatureSizeHint());
    }
    
    /**
     * Override this class to generate a new module from command line arguments.  It should return null if the modName is not correct for this module.
     *
     * @return the new module generated from the specified switches, or null if the modName is not appropriate for this module.
     * @param modName the string following tthe -mod switch
     * @param switchMap the switches provided at the command line (i.e. "-file") mapped to the index in the string array. Any switches that are processed should be removed.
     * @param args the arguments array provided at the command line.
     */
    public static SDDModule makeModuleFromArgs(String modName, java.util.HashMap<String,Integer> switchMap, String[] args, StereoCamera camera) throws  ParsingException{
        return null;
    }
    
    class SDDMKeyListener extends KeyAdapter {
        private StereoCameraControlFrame frame;
        public SDDMKeyListener( StereoCameraControlFrame frame ) {
            super();
            this.frame = frame;
        }
        
        public void keyTyped(KeyEvent e) {
            char t = e.getKeyChar();
            switch  (e.getKeyChar()) {
                
                case 'f':
                case 'F':
                    javax.swing.JFrame[] jf = frame.getAttachedFrames();
                    for (javax.swing.JFrame f : jf) {
                        if (f instanceof GLJFrame) {
                            GLJFrame gljf = (GLJFrame)f;
                            boolean fs = gljf.isFullscreen();
                            gljf.setFullscreen(!fs);
                        }
                    }
                    break;
                case 'w':
                    cam.translate(frame.getControlPanel().getTransStep());
                    frame.redisplay();
                    break;
                case 'S':
                    //take screenshot
                    javax.swing.JFileChooser jfc = new javax.swing.JFileChooser();
                    jfc.setMultiSelectionEnabled(false);
                    jfc.setDialogType(jfc.SAVE_DIALOG);
                    jfc.setFileFilter(new ImageIOFileFilter(false,true));
                    jfc.setDialogTitle("Location to Save Screenshot");
                    
                    
                    if (jfc.APPROVE_OPTION == jfc.showSaveDialog(null)) {
                        File f = jfc.getSelectedFile();
                        try {
                            this.frame.takeScreenshotFromFrames(f);
                        } catch (IOException ex) {
                            if(ex.getMessage().equals("File name does not have extension"))
                                javax.swing.JOptionPane.showMessageDialog(null, "Need to provide extension. Screenshot not taken.","Screenshot error",javax.swing.JOptionPane.ERROR_MESSAGE);
                            else {
                                Writer result = new StringWriter();
                                PrintWriter printWriter = new PrintWriter(result);
                                javax.swing.JOptionPane.showMessageDialog(null,"Error taking screenshot:\n"+result.toString(),"Screenshot error",javax.swing.JOptionPane.ERROR_MESSAGE);
                            }
                        }
                    } else
                        javax.swing.JOptionPane.showMessageDialog(null,"No file selected, screenshot not saved.","Screenshot Save Error",javax.swing.JOptionPane.ERROR_MESSAGE);
                    break;
                case 's':
                    cam.translate(-frame.getControlPanel().getTransStep());
                    frame.redisplay();
                    break;
                case 'a':
                    cam.strafe(-frame.getControlPanel().getTransStep());
                    frame.redisplay();
                    break;
                case 'd':
                    cam.strafe(frame.getControlPanel().getTransStep());
                    frame.redisplay();
                    break;
                case 'q':
                case 'Q':
                    if (e.isControlDown()||e.isAltDown()||e.isMetaDown()||e.isShiftDown())
                        System.exit(0);
                    cam.roll(-frame.getControlPanel().getRotStep());
                    frame.redisplay();
                    break;
                case 'e':
                    cam.roll(frame.getControlPanel().getRotStep());
                    frame.redisplay();
                    break;
            }
        }
    }
    
    class SDDMMouseListener implements MouseListener, MouseMotionListener, MouseWheelListener {
        private StereoCameraControlFrame frame;
        private int oldMouseX,oldMouseY;
        private int mouseMoving; //if negative, mouse not moving
        public SDDMMouseListener( StereoCameraControlFrame frame) {
            this.frame = frame;
            oldMouseX = 0;
            oldMouseY = 0;
            mouseMoving = -1;
        }
        public void mouseClicked(MouseEvent e) {
            //Toggle Target Sphere on
            if(e.getButton()==e.BUTTON2) {
                cam.drawSphereOnUpdate = cam.drawSphereOnUpdate?false:true;
                frame.getControlPanel().setTargetInteractionMode(cam.drawSphereOnUpdate);
                frame.redisplay();
            }
        }
        public void mouseEntered(MouseEvent e) {
            //Only in for interface compatibility
        }
        public void mouseExited(MouseEvent e) {
            //Only in for interface compatibility
        }
        public void mousePressed(MouseEvent e) {
            oldMouseX = e.getX();
            oldMouseY = e.getY();
            mouseMoving = e.getButton();
        }
        public void mouseReleased(MouseEvent e) {
            mouseMoving = -1;
        }
        public void mouseDragged(MouseEvent e) {
            int deltaX = e.getX() - oldMouseX;
            deltaX *= frame.getControlPanel().getMouseSens();
            int deltaY = e.getY() - oldMouseY;
            deltaY *= frame.getControlPanel().getMouseSens();
            
            switch(mouseMoving){
                case MouseEvent.BUTTON3:
                    if (frame.getControlPanel().isTargetInteractionMode()) {
                        float dist = cam.getTargetDistance();
                        if(deltaY>3)
                            deltaY -= 3;
                        else if (deltaY<-3)
                            deltaY += 3;
                        else
                            deltaY=0;
                        float offset = (deltaX-10*deltaY)*getMinFeatureSizeHint();
                        float newD = dist+offset;
                        cam.setTargetDistance((newD > 0) ? newD : getMinFeatureSizeHint());
                        this.frame.getControlPanel().cameraSettingsToPanel();
                    } else {
                        if (deltaX!=0)
                            cam.strafe(0.1f*deltaX);
                        if (deltaY!=0)
                            cam.slew(-0.1f*deltaY);
                    }
                    break;
                case MouseEvent.BUTTON1:
                    if (frame.getControlPanel().isTargetInteractionMode()) {
                        if (deltaX!=0)
                            cam.rotateAboutTargetPoint(deltaX);
                        if (deltaY!=0)
                            cam.tiltAboutTargetPoint(deltaY);
                    } else {
                        if (deltaX!=0)
                            cam.yaw(-deltaX);
                        if (deltaY!=0)
                            cam.pitch(-deltaY);
                    }
            }
            
            oldMouseX = e.getX();
            oldMouseY = e.getY();
            frame.redisplay();
        }
        public void mouseMoved(MouseEvent e) {
            //Only in for interface compatibility
        }
        public void mouseWheelMoved(MouseWheelEvent e) {
            float dist = cam.getTargetDistance();
            dist -= e.getUnitsToScroll()*frame.getControlPanel().getMouseWheelSens();
            if (dist > 0) {
                cam.setTargetDistance(dist);
                this.frame.getControlPanel().cameraSettingsToPanel();
                frame.redisplay();
            }
        }
        
    }
    
}

