/*
 * HeightField.java
 *
 * Created on May 19, 2006, 12:22 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package sddm.data;

import java.awt.image.*;
import java.io.*;
import java.util.*;
import javax.vecmath.*;
import javax.imageio.*;
import sddm.*;

/**
 * This class represents a two-dimensional scalar field storing float values.
 * @author Erik
 */
public class HeightField {
    private float[][] data;
    private float maxHeight;
    private float minHeight;
    private float avgHeight;
    //for computation of normals
    private Vector3f[][] normals;
    private Vector3f[] normalCompVecs;
    private int normalGranularity = 1;
    private float normalScaling = 1.0f;
    
    /**
     * Used to specify file type in loading a {@link #sddm.data.HeightField HeightField}
     */
    public enum hfFileType {ASCII_1D, ASCII_2D, IMAGE};
    
    
    /**
     * Creates a new instance of HeightField with a single value of 0
     */
    public HeightField() {
        this(1,1);
    }
    
    /**
     * Creates a new instance of HeightField of size (xSize, ySize) with all values of 0
     * @param xSize the number of rows
     * @param ySize the number of columns
     */
    public HeightField(int xSize, int ySize) {
        data = new float[xSize][ySize];
        for(int i = 0; i < xSize; ++i) {
            for(int j = 0; j < ySize; ++j) {
                data[i][j] = 0;
            }
        }
        
        maxHeight = minHeight = avgHeight = 0;
        
    }
    
    /**
     * Creates a new instance of HeightField from the specified 2D array.
     * @param dataIn Array to generate HeightField from.  Note that this array <B>must</B> be square (i.e. all the sub-arrays must be the same length).
     */
    public HeightField(float[][] dataIn) {
        setData(dataIn);
    }
    /**
     * Creates a new instance of HeightField from the specified file.
     * @param f The file to load the HeightField from.
     * @param t The format of the input File.
     * @throws sddm.FileFormatException If the file is not the appropriate size (see setData(File, hfFileType))
     * @throws java.io.IOException if there is a problem accessing the file
     */
    public HeightField(File f, hfFileType t) throws IOException, FileFormatException {
        setData(f, t);
    }
    
    /**
     * Creates a new instance of HeightField from an array. 
     * 
     * @param dataIn the array with the data in row-major(?) order
     * @param xSizeIn the number of rows
     * @param ySizeIn the number of columns
     * @throws sddm.SDDMException if the data array is the incorrect length
     */
    public HeightField(float[] dataIn, int xSizeIn, int ySizeIn) throws SDDMException {
        setData(dataIn,xSizeIn,ySizeIn);
    }
    
    /**
     * Creates a new instance of HeightField by copying the data from another HeightField. Does NOT copy cached normals
     */
    public HeightField(HeightField fieldToCopy) {
        int xSize = fieldToCopy.getXSize();
        int ySize = fieldToCopy.getYSize();
        this.data = new float[xSize][ySize];
        for (int i = 0; i < xSize; i++) {
            for (int j = 0; j < ySize; j++) {
                this.data[i][j] = fieldToCopy.data[i][j];
            }
            
        }
        this.avgHeight = fieldToCopy.avgHeight;
        this.maxHeight = fieldToCopy.maxHeight;
        this.minHeight = fieldToCopy.minHeight;
        this.normalGranularity = fieldToCopy.normalGranularity;
        this.normalScaling = fieldToCopy.normalScaling;
    }
    
    
    /**
     * Returns the size of the first dimension of the HeightField.
     * @return The length of the HeightField
     */
    public int getXSize() {
        return data.length;
    }
    /**
     * Returns the size of the second dimension of the HeightField.
     * @return The width of the HeightField
     */
    public int getYSize() {
        return data[0].length;
    }
    /**
     * Returns the largest value in the HeightField.
     * @return The largest value in the HeightField
     */
    public float getMaxValue() {
        return maxHeight;
    }
    
    /**
     * Returns the smallest value in the HeightField.
     * @return The smallest value in the HeightField.
     */
    public float getMinValue() {
        return minHeight;
    }
    
    /**
     * The average value determined from all the heightField values.
     * @return The average value
     */
    public float getAvgValue() {
        return avgHeight;
    }
    /**
     * The factor to multiply the values by to fit all the values into a box with a height equaly to the average of the length and the width divided by 3 (z=x+y/6).
     * @return the multiplicative factor (i.e. use getDataPoint(x,y)*getZScaleFactor() )
     */
    public float getZScaleFactor() {
        float targetSize = (this.getXSize()+this.getYSize())/6;
        float largestProjection = Math.max(this.maxHeight,this.minHeight);
        return targetSize/largestProjection/2;
    }
    
    /**
     * Changes the data in this HeightField based on a 2D array.
     * @param data An array of new data to use.
     */
    public void setData(float[][] data) {
        //want deep copy - clone only does copy of first level...
        for(int i = 0; i < data.length; ++i)
            this.data[i] = data[i].clone();
        fixStats();
    }
    
    /**
     *  Changes the data in this HeightField from a float array
     * 
     * @param data a list of elements in row-major(??) order (i.e. a row of constant y =0, then a row of y=1, etc.)
     * @param xSize the width of the data set
     * @param ySize the height of the data set
     * @throws sddm.SDDMException if the data array is the incorrect length
     */
    public void setData(float[] data, int xSize, int ySize) throws SDDMException {
        //TODO: test
        if (data.length != (xSize*ySize))
            throw new SDDMException("data array is not xSize*ySize");
        /*int num = 0;
        for(int i = 0; i < xSize; ++i) {
            for (int j = 0; j < ySize; ++j) 
                this.data[i][j] = data[num++];
        }*/
        int i = 0;
        int j = 0;
        for(int num = 0; num < data.length; ++num) {
            this.data[i++][j] = data[num];
            if (i >= xSize)
                j++;
        }
    }
    
    /**
     * Changes the data in the HeightField based on a file.
     * @param f The file to load the HeightField from.
     * @param t The format of the input File.
     * @throws sddm.FileFormatException If the data in the file does not match the expected file format.
     * @throws java.io.IOException if there is a problem accessing the file
     */
    public void setData(File f, hfFileType t) throws IOException, FileFormatException {
        int xSize, ySize;
        
        switch (t) {
            case IMAGE:
                BufferedImage bi = ImageIO.read(f);
                xSize = bi.getWidth();
                ySize = bi.getHeight();
                data = new float[xSize][ySize];
                //Iterate over pixels and set data value to "average" color (i.e. greyscale equivalent value)
                for(int i=0;i<xSize;++i) {
                    for(int j=0;j<ySize;++j) {
                        int argb = bi.getRGB(i,j);
                        int red = (argb >> 16) & 0xff;
                        int green = (argb >> 8) & 0xff;
                        int blue = argb & 0xff;
                        //flip the y-axis because the image coordinate flipped from openGL/cartesian coordinates
                        data[i][ySize-j-1] = (red+green+blue)/3;
                    }
                }
                break;
            case ASCII_1D:
                //TODO:fix error checking
                Scanner s = new Scanner(f);
                int i,j;
                if (s.hasNextFloat())
                    xSize = (int)s.nextFloat();
                else
                    throw new FileFormatException("File does not include Height Field x size");
                
                if (s.hasNextFloat())
                    ySize = (int)s.nextFloat();
                else
                    throw new FileFormatException("File does not include Height Field y size");
                
                data = new float[xSize][ySize];
                i = 0;
                j = 0;
                while (s.hasNextFloat()) {
                    data[i][j] = s.nextFloat();
                    if(++i >= xSize) {
                        i=0;
                        if(++j >= ySize && s.hasNextFloat())
                            throw new FileFormatException("Too many numbers in file");
                    }
                }
                if (i!=0||j!=ySize)
                    throw new FileFormatException("Too few numbers in file");
                break;
            case ASCII_2D:
                //Find number of Lines and number of elements in first line (x and y size)
                BufferedReader reader = new BufferedReader(new FileReader(f));
                StringTokenizer strTok;
                String currLine = reader.readLine();
                strTok = new StringTokenizer(currLine);
                xSize = strTok.countTokens();
                ySize = 1;
                while(reader.readLine()!=null)
                    ++ySize;
                data = new float[xSize][ySize];
                //Reset reader to start of file
                reader = new BufferedReader(new FileReader(f));
                i=j=0;
                //read first line (although as its already been read ... eh)
                currLine = reader.readLine();
                //loop over lines
                while (currLine!=null) {
                    strTok = new StringTokenizer(currLine);
                    while (strTok.hasMoreTokens()) {
                        float fl = Float.parseFloat(strTok.nextToken());
                        data[i][j] = fl;
                        ++i;
                    }
                    if (i!=xSize)
                        throw new FileFormatException("Wrong number of values in line #"+j+1);
                    i=0;
                    ++j;
                    currLine = reader.readLine();
                }
                break;
        }
        fixStats();
    }
    
    public void saveData(File f, hfFileType t) throws SDDMException, IOException {
        this.saveData(f,t,this.getMaxValue(),this.getMinValue());
        
    }
    
    public void saveData(File f, hfFileType t, float max, float min) throws SDDMException, IOException {
        if(f.exists()) {
            if(!f.delete())
                throw new IOException("couldn't delete file while saving HeightField");
        }
        if(!f.createNewFile())
            throw new IOException("couldn't create file while saving HeightField");
        
        switch (t) {
            case IMAGE:
                String[] fn = f.getName().split("\\.");
                String ext = fn[fn.length-1];
                //Iterator<ImageWriter> writers = ImageIO.getImageWritersBySuffix(ext);
                //if (!writers.hasNext())
                //    throw new SDDMException("Image specified by extension not supported.");
                
                BufferedImage im = new BufferedImage(this.getXSize(),this.getYSize(),BufferedImage.TYPE_INT_RGB);
                float factor = 255.0f/(max - min);
                
                for (int i = 0; i < this.getXSize(); i++) {
                    for (int j = 0; j < this.getYSize(); j++) {
                        //TODO implement better - use grayscale values instead of just 0-255
                        int rgb = (int)((data[i][j]-min)*factor);
                        rgb += (rgb << 8) + (rgb << 16);
                        //flipping j to switch from OpenGL to Image coordinates
                        im.setRGB(i,this.getYSize()-j-1,rgb);
                    }
                }
                if(!ImageIO.write(im,ext,f))
                    throw new SDDMException("Image specified by extension not supported.");
                
                break;
            case ASCII_1D:
                PrintStream fs = new PrintStream(f);
                
                fs.print(this.getXSize());
                fs.print(' ');
                fs.print(this.getYSize());
                fs.print(' ');
                for (int i = 0; i < this.getXSize(); i++) {
                    for (int j = 0; j < this.getYSize(); j++) {
                        fs.print(data[i][j]);
                        fs.print(' ');
                    }
                }
                fs.close();
                break;
            case ASCII_2D:
                fs = new PrintStream(f);
                
                for (int i = 0; i < this.getXSize(); i++) {
                    for (int j = 0; j < this.getYSize(); j++) {
                        fs.print(data[i][j]);
                        fs.print(' ');
                    }
                    fs.println();
                }
                fs.close();
                break;
        }
    }
    
    /**
     * Retrieves a value from the HeightField
     * @param xVal The x-coordinate
     * @param yVal The y-coordinate
     * @return The value at (x,y)
     */
    public float getData(int xVal, int yVal) {
        return data[xVal][yVal];
    }
    
    /**
     * Generates or retrieves the vector normal to the height field surface at the 
     * specified point. This calculation depends on the normalScaling and 
     * normalGranularity properties. 
     * Note that once a particular vector is computed once, it only needs to be 
     * retrieved, which takes much less time.
     * @param xVal The x-coordinate
     * @param yVal The y-coordinate
     * @return the Normal Vector at (xVal,yVal)
     */
    public Vector3f getNormal(int xVal, int yVal) {
        if (this.normals == null)
            allocateNormals();
        
        if (this.normals[xVal][yVal] == null)
            this.normals[xVal][yVal] = new Vector3f(computeNormal(xVal, yVal));
        
        return normals[xVal][yVal];
        
    }
    
    /**
     * Retrieves an array storing all the data.  Note that this returns a reference (for performance reasons) - Altering this data is not recommended.
     * @return A reference to the array storing the values.  It is highly recommended that this data not be altered.
     */
    public float[][] getData() {
        return data;
    }
    
    /**
     * Specifies the scaling for computation of the HeightField normals.
     * This resets the Normals, however, so they all need to be re-computed.
     * @param normalScaling the value to multiply the z-coordinate of the vectors crossed to generate the normal vector.
     */
    public void setNormalScaling(float normalScaling) {
        clearNormals();
        this.normalScaling = normalScaling;
    }
    
    /**
     * Retrieves the z-coordinate scale used to compute the normals.
     * @return the scaling factor
     */
    public float getNormalScaling() {
        return normalScaling;
    }
    
    /**
     * Specifies the granularity used to compute the normals in this HeightField, 
     * so they must be re-computed following a change to the granularity. 
     * @param normalGranularity the granularity to use (i.e. two is every-other vertex, 3 is every third, etc.) should be on [0,maxSize]
     */
    public void setNormalGranularity(int normalGranularity) {
        clearNormals();
        this.normalGranularity = normalGranularity;
    }
    
    /**
     * Retrieves the granularity used to calculate the normals on this HeightField.
     * @return the granularity (1 is all, 2 is every other, 3 is every third, etc.)
     */
    public int getNormalGranularity() {
        return normalGranularity;
    }
    
    /**
     * Copies the array storing all the data.
     * @return A new copy of the data in the HeightField.
     */
    public float[][] copyData() {
        int xSize = this.getXSize();
        int ySize = this.getYSize();
        float[][] copy = new float[xSize][ySize];
        for (int i = 0; i < xSize;++i)
            copy[i] = this.data[i].clone();
        return copy;
    }
    
    
    //---------------------Opeartions that change Height Field----------------------------------
    /**
     * Smooths the values in this height field using nearest-neighbor convolution. <br>
     * e.g. newZ = (currZ+weight*leftZ+weight*rightZ+weight*aboveZ+weight*belowZ)/(1+4*weight)
     * @param corners includes diagonal neighbors if true
     * @param neighborWeight the weight for the convolution elements relative to the weight of the element itself
     */
    public void smoothConvolution(float neighborWeight, boolean corners) {
        int xmo = this.getXSize()-1;
        int ymo = this.getYSize()-1;
        
        if(corners) {
            float divisor = 1+8*neighborWeight;
            
            for(int i = 1; i < xmo; i++) {
                for(int j = 1; j < ymo;j++) {
                    data[i][j] = (data[i][j]+neighborWeight*(data[i+1][j]+data[i][j+1]+data[i-1][j]+data[i][j-1]+data[i+1][j+1]+data[i+1][j-1]+data[i-1][j+1]+data[i-1][j-1])) / divisor;
                }
            }
        }
        else {
            float divisor = 1+4*neighborWeight;
            
            for(int i = 1; i < xmo; i++) {
                for(int j = 1; j < ymo;j++) {
                    data[i][j] = (data[i][j]+neighborWeight*(data[i+1][j]+data[i][j+1]+data[i-1][j]+data[i][j-1])) / divisor;
                }
            }
        }
        clearNormals();
        fixStats();
    }
    
    /**
     * Move the height of the field uniformly
     * @param amount the translation amount
     */
    public void zTranslate(float amount) {
        for(int i = 0; i < this.getXSize(); i++) {
            for(int j = 0; j < this.getYSize();j++) {
                data[i][j] += amount;
            }
        }
        fixStats();
    }

    /**
     * Sets all elements of this height field above a certain value  to that value.
     * @param height the z-value at which to cap the height field.
     */
    public void cap(float height) {
        for(int i = 0; i < this.getXSize(); i++) {
            for(int j = 0; j < this.getYSize();j++) {
                if (data[i][j] >= height) {
                    data[i][j] = height;
                    normals[i][j] = null;
                }
            }
        }
        fixStats();
    }
    
    /**
     * Sets all elements of this height field below a certain value to that value.
     * @param height the z-value at which to flood the height field.
     */
    public void flood(float height) {
        for(int i = 0; i < this.getXSize(); i++) {
            for(int j = 0; j < this.getYSize();j++) {
                if (data[i][j] < height) {
                    data[i][j] = height;
                    normals[i][j] = null;
                }
            }
        }
        fixStats();
    }
    
    public void fractalFault() {
        //TODO:implement
    }
    
    public void fft() {
        HeightField imag = new HeightField(this.getXSize(),this.getYSize());
        fft(false,this,imag);
        this.magnitude(imag);
    }
    
    public void inverseFft() {
        HeightField imag = new HeightField(this.getXSize(),this.getYSize());
        fft(true,this,imag);
        this.magnitude(imag);
    }
    
    public void oneOverFFilter(float beta) {
        //TODO: test
        int xSize = this.getXSize();
        int ySize = this.getYSize();
        
        for (int i = 0; i < xSize; i++) {
            for (int j = 0; j < ySize; j++) {
                float xDist = i - xSize/2.0f;
                float yDist = j - ySize/2.0f;
                //set the center value to the average, because 1/f is undefined there, but for fourier transformss its correct
                if (xDist == 0 && yDist == 0)
                    data[i][j] = this.avgHeight;
                else {
                    float distSquared = xDist*xDist+yDist*yDist;
                    data[i][j] = data[i][j]/(float)Math.pow(distSquared,beta/2.0);
                }
            }
        }
    }
    
    
    public static HeightField generateRandomField(int xSize, int ySize) {
        java.util.Random rand = new java.util.Random();
        float[] data = new float[xSize*ySize];
        
        for (int i = 0; i < data.length; ++i) 
            data[i] = rand.nextFloat();
        
        return new HeightField(data,xSize,ySize);
    }
    
    public static HeightField generatePoweredSpectralField(int xSize, int ySize, float beta) {
        HeightField real = HeightField.generateRandomField(xSize,ySize);
        HeightField imag = new HeightField(xSize,ySize);
        fft(false,real,imag);
        real.oneOverFFilter(beta);
        imag.oneOverFFilter(beta);
        fft(true,real,imag);
        real.magnitude(imag);
        return real;
        //TODO: test: http://astronomy.swin.edu.au/~pbourke/modelling/freqland/, http://astronomy.swin.edu.au/~pbourke/fractals/noise/
        
    }
    
    public static void fft(HeightField real, HeightField imag) {
        fft(false,real,imag);
    }
    
    public static void inverseFft(HeightField real, HeightField imag) {
        fft(true,real,imag);
    }
    
    
    //--------------------------private internals below here -------------------
    private void magnitude(HeightField otherField) {
        for (int i = 0; i < this.getXSize(); i++) {
            for (int j = 0; j < this.getYSize(); j++) {
                float thisVal = this.data[i][j]*this.data[i][j];
                float otherVal = otherField.data[i][j];
                thisVal = thisVal*thisVal  + otherVal*otherVal;
                this.data[i][j] = (float)Math.sqrt(thisVal);
            }
        }
    }
    
    private static void fft(boolean inverse, HeightField real, HeightField imag) {
        if (isPowerOfTwo(real.getXSize())&&isPowerOfTwo(real.getXSize())) {
            //TODO:implement
        }
        else {
            //neeed to find a generic non radix-2 FFT
            throw new SDDMException("tried to fft a field that isn't power-of-2 sized");
        }
    }
    
    private static boolean isPowerOfTwo(int num) {
        int tester = 1;
        while(num > tester) 
            tester*=2;
        return (tester==num);
    }
    
    private void fixStats() {
        maxHeight = Float.MIN_VALUE;
        minHeight = Float.MAX_VALUE;
        avgHeight = 0;
        int xSize = this.getXSize();
        int ySize = this.getYSize();
        for(int i=0;i<xSize;++i) {
            for(int j=0;j<ySize;++j) {
                avgHeight+=data[i][j];
                if (maxHeight < data[i][j])
                    maxHeight = data[i][j];
                if (minHeight > data[i][j])
                    minHeight = data[i][j];
            }
        }
        avgHeight = avgHeight / (xSize*ySize);
    }
    
    private void allocateNormals() {
        this.normalCompVecs = new Vector3f[9];
        for (int i = 0; i < this.normalCompVecs.length; i++)
            this.normalCompVecs[i] = new Vector3f();
        
        this.normals = new Vector3f[this.getXSize()][this.getYSize()];
    }
    
    private void clearNormals() {
        if (this.normals == null)
            allocateNormals();
        
        for(Vector3f[] va : this.normals) {
            for (int i = 0; i < va.length; i++)
                va[i] = null;
        }
    }
    
    /**
     * Computes the normal at the grid point (xCoord, yCoord) and assigns it to the first vector in normalCompVecs
     * @param xCoord the column of the grid point (strips iterate over these)
     * @param yCoord the row of the grid point (one strip each)
     * @return the calculated normal vector.  Note that this vector may change - make a new one and initialize it to store this normal
     */
    private Vector3f computeNormal(int xCoord,int yCoord) {
        //uses all of the vectors past the first for computing the normal
        Vector3f[] vecs = this.normalCompVecs;
        int gran = this.normalGranularity;
        float scal = this.normalScaling;
        
        float value = getData(xCoord,yCoord);
        if ((vecs.length - 1) == 8) {
            
            int vecsToCross = 1;
            int xp1 = xCoord + 1;
            int xm1 = xCoord - 1;
            int yp1 = yCoord + 1;
            int ym1 = yCoord - 1;
            //check edges
            boolean left = (xCoord == 0);
            boolean right = (xCoord >= (getXSize() - gran));
            boolean bottom = (yCoord == 0);
            boolean top = (yCoord >= (getYSize() - gran));
            
            if (left&&bottom) {
                //bottom-left corner
                vecs[1].set(gran,0,scal*(getData(xp1,yCoord)-value));
                vecs[2].set(gran,gran,scal*(getData(xp1,yp1)-value));
                vecs[3].set(0,gran,scal*(getData(xCoord,yp1)-value));
                vecsToCross = 3;
            } else if (left&&top) {
                //upper-left corner
                vecs[1].set(0,-gran,scal*(getData(xCoord,ym1)-value));
                vecs[2].set(gran,-gran,scal*(getData(xp1,ym1)-value));
                vecs[3].set(gran,0,scal*(getData(xp1,yCoord)-value));
                vecsToCross = 3;
            } else if (right&&bottom) {
                //bottom-right
                vecs[1].set(0,gran,scal*(getData(xCoord,yp1)-value));
                vecs[2].set(-gran,gran,scal*(getData(xm1,yp1)-value));
                vecs[3].set(-gran,0,scal*(getData(xm1,yCoord)-value));
                vecsToCross = 3;
            } else if (right&&top) {
                //upper-right
                vecs[1].set(-gran,0,scal*(getData(xm1,yCoord)-value));
                vecs[2].set(-gran,-gran,scal*(getData(xm1,ym1)-value));
                vecs[3].set(0,-gran,scal*(getData(xCoord,ym1)-value));
                vecsToCross = 3;
            } else if (left) {
                //left edge
                vecs[1].set(0,-gran,scal*(getData(xCoord,ym1)-value));
                vecs[2].set(gran,-gran,scal*(getData(xp1,ym1)-value));
                vecs[3].set(gran,0,scal*(getData(xp1,yCoord)-value));
                vecs[4].set(gran,gran,scal*(getData(xp1,yp1)-value));
                vecs[5].set(0,gran,scal*(getData(xCoord,yp1)-value));
                vecsToCross = 5;
            } else if (right) {
                //right edge
                vecs[1].set(0,gran,scal*(getData(xCoord,yp1)-value));
                vecs[2].set(-gran,gran,scal*(getData(xm1,yp1)-value));
                vecs[3].set(-gran,0,scal*(getData(xm1,yCoord)-value));
                vecs[4].set(-gran,-gran,scal*(getData(xm1,ym1)-value));
                vecs[5].set(0,-gran,scal*(getData(xCoord,ym1)-value));
                vecsToCross = 5;
            } else if (top) {
                //top edge
                vecs[1].set(-gran,gran,scal*(getData(xm1,yCoord)-value));
                vecs[2].set(-gran,-gran,scal*(getData(xm1,ym1)-value));
                vecs[3].set(0,-gran,scal*(getData(xCoord,ym1)-value));
                vecs[4].set(gran,-gran,scal*(getData(xp1,ym1)-value));
                vecs[5].set(gran,0,scal*(getData(xp1,yCoord)-value));
                vecsToCross = 5;
            } else if (bottom) {
                //bottom edge
                vecs[1].set(gran,0,scal*(getData(xp1,yCoord)-value));
                vecs[2].set(gran,gran,scal*(getData(xp1,yp1)-value));
                vecs[3].set(0,gran,scal*(getData(xCoord,yp1)-value));
                vecs[4].set(-gran,gran,scal*(getData(xm1,yp1)-value));
                vecs[5].set(-gran,0,scal*(getData(xm1,yCoord)-value));
                vecsToCross = 5;
            } else {
                //clockwise from right
                vecs[1].set(gran,0,scal*(getData(xp1,yCoord)-value));
                vecs[2].set(gran,gran,scal*(getData(xp1,yp1)-value));
                vecs[3].set(0,gran,scal*(getData(xCoord,yp1)-value));
                vecs[4].set(-gran,gran,scal*(getData(xm1,yp1)-value));
                vecs[5].set(-gran,0,scal*(getData(xm1,yCoord)-value));
                vecs[6].set(-gran,-gran,scal*(getData(xm1,ym1)-value));
                vecs[7].set(0,-gran,scal*(getData(xCoord,ym1)-value));
                vecs[8].set(gran,-gran,scal*(getData(xp1,ym1)-value));
                vecsToCross = 8;
            }
            
            for(int i = 1; i < vecsToCross; ++i)
                vecs[i].cross(vecs[i],vecs[i+1]);
            if (vecsToCross == 8) {
                vecs[8].cross(vecs[8],vecs[1]);
                vecsToCross = 9;
            }
            vecs[0].set(0,0,0);
            for(int i = 1; i < (vecsToCross - 1); ++i)
                vecs[0].add(vecs[i]);
            
        } else
            throw new SDDMException("Invalid number of vectors for normal calculatuion");
        
        vecs[0].normalize();
        return vecs[0];
    }
}
