/*
 * ParametricModule.java
 * -laf native -mod ps -ls 1 0 0 2.667 30 10 0.01 1000 -satgradient 0.8 0 0.8
 */

package sddm.modules;

import java.awt.event.KeyEvent;
import sddm.*;
import sddm.data.*;
import javax.media.opengl.*;

/**
 *
 * @author Erik J. Tollerud
 */
public class ParametricModule extends PointDrawingModule {
    private ParametricSystem sys;
    private int iterateOnRedisplay;
    
    /** Creates a new instance of ParametricModule */
    public ParametricModule(ParametricSystem sys) {
        super(sys.getPointList());
        joined = true;
        this.sys = sys;
        iterateOnRedisplay = 0;
    }
    
    /** Creates a new instance of ParametricModule */
    public ParametricModule(ParametricSystem sys, StereoCamera camera) {
        super(sys.getPointList(),camera);
        joined = true;
        this.sys = sys;
        iterateOnRedisplay = 0;
    }
    
    public void setIterateOnRedisplay(int numOfItersPerRedisplay) {
        if (numOfItersPerRedisplay >= 0)
            this.iterateOnRedisplay = numOfItersPerRedisplay;
        else
            this.iterateOnRedisplay = 0;
    }
    
    public int getIterateOnRedisplay() {
        return this.iterateOnRedisplay;
    }
    
    public SDDModuleControlPanel newModuleControlPanel() {
        return new ParametricModuleControlPanel(this);
    }
    
    public static ParametricModuleSetupPanel newStartupPanel() {
        return new ParametricModuleSetupPanel();
    }
    
    public static SDDModule makeModuleFromArgs(String modName, java.util.HashMap<String,Integer> switchMap, String[] args, StereoCamera camera) throws  ParsingException{
        //accepted module names: pm|ps|parametric|parametricsystem|parametricmodule
        //switches processed:
        // -ls x0 y0 z0 b r s deltat iters 
        // -rm x0 y0 z0 a b c deltat iters 
        //TODO:add diffeq processor
        //[coloring switches]
        
        if(modName.matches("pm|ps|parametric|parametricsystem|parametricmodule")) {
            ParametricModule m = null;
            
            Integer lsIndx = switchMap.remove("-ls");
            if (lsIndx != null) {
                try {
                float x0 = Float.parseFloat(args[lsIndx+1]);
                float y0 = Float.parseFloat(args[lsIndx+2]);
                float z0 = Float.parseFloat(args[lsIndx+3]);
                float b = Float.parseFloat(args[lsIndx+4]);
                float r = Float.parseFloat(args[lsIndx+5]);
                float s = Float.parseFloat(args[lsIndx+6]);
                float deltat = Float.parseFloat(args[lsIndx+7]);
                int iters = Integer.parseInt(args[lsIndx+8]);
                
                
                UniformParametricGenerator.Generator gen = new ParametricLSTypePanel.LSEqnGenerator(b,r,s);
                ParametricSystem ps = new ParametricSystem(new UniformParametricGenerator(deltat,gen,x0,y0,z0,0.0f));
                ps.iterate(iters);
                
                m = new ParametricModule(ps,camera);
                }
                catch (Exception e) {
                    throw new ParsingException("Problem recognizing Lorenz-Salzman settings while generating Parametric Module:"+e.getMessage(),e);
                }
            }
            
            Integer rmIndx = switchMap.remove("-rm");
            if (rmIndx != null) {
                try {
                float x0 = Float.parseFloat(args[rmIndx+1]);
                float y0 = Float.parseFloat(args[rmIndx+2]);
                float z0 = Float.parseFloat(args[rmIndx+3]);
                float a = Float.parseFloat(args[rmIndx+4]);
                float b = Float.parseFloat(args[rmIndx+5]);
                float c = Float.parseFloat(args[rmIndx+6]);
                float deltat = Float.parseFloat(args[rmIndx+7]);
                int iters = Integer.parseInt(args[rmIndx+8]);
                
                
                UniformParametricGenerator.Generator gen = new ParametricRosslerTypePanel.RosslerGenerator(a,b,c);
                ParametricSystem ps = new ParametricSystem(new UniformParametricGenerator(deltat,gen,x0,y0,z0,0.0f));
                ps.iterate(iters);
                
                m = new ParametricModule(ps,camera);
                }
                catch (Exception e) {
                    throw new ParsingException("Problem recognizing Rossler Map settings while generating Parametric Module:"+e.getMessage(),e);
                }
            }
            
            if (m == null)
                m = new ParametricModule(new ParametricSystem(),camera);
            
            Coloring c = Coloring.makeColoringFromArgs(switchMap,args);
            if (c != null)
                m.setColorer(c);
            return m;
        }
        return null;
    }
    public void setSystem(ParametricSystem sys) {
        this.sys = sys;
        this.setPoints(sys.getPointList());
        this.recompileDisplayList();
    }
        
    public ParametricSystem getSystem() {
        return sys;
    }
    
    public static String moduleTitle() {
        return "Parametric Module";
    }
    
    public void preDraw(GL gl) {
        super.preDraw(gl);
        if (this.iterateOnRedisplay > 0) {
            sys.iterate(this.iterateOnRedisplay);
            this.recompileDisplayList();
        }
    }
    
    public int iterPerKey = 1;
    public void attachEventListeners(StereoCameraControlFrame frame) {
        super.attachEventListeners(frame);
        final StereoCameraControlFrame frameF = frame;
        frame.addKeyListenerToJFrames(new java.awt.event.KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                if(e.getKeyChar() == 'I') {
                    getSystem().iterate(iterPerKey);
                    recompileDisplayList();
                    frameF.redisplay();
                }
            }
        
        });
    }
}
